import os
import os.path
import numpy as np
import numpy.random
import torch
import torch.utils.data as udata
import random
from util.util import mkdir, mkdirs
import fnmatch
import random
from data.base_dataset import BaseDataset, get_params, get_transform
from data.image_folder import make_dataset
from PIL import Image

from library import *

class infer4recDataset(BaseDataset):
    """A dataset class for testing! It will provide (in a serialize way) entire images to denoise.
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new dataset-specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can use this flag to add training-specific or test-specific options.

        Returns:
            the modified parser.
        """
        BaseDataset.modify_commandline_options(parser, is_train)
        parser.add_argument('--crop_data', type=str, default=None,
                            help='Crops all output data from this dataset. --crop_data x,y does img[:x,:y].')
        return parser

    def __init__(self, opt):
        """Initialize this dataset class.

        Parameters:
            opt (Option class) -- stores all the experiment flags; needs to be a subclass of BaseOptions
        """
        BaseDataset.__init__(self, opt)

        self.T, self.iT = define_transforms(opt)
        self.rootdir = opt.dataroot

        self.gt_paths = os.path.join(self.rootdir, opt.gtFolder)
        self.n_paths = os.path.join(self.rootdir, opt.nFolder)

        if opt.rawFolder in ["\noraw", " ", None]:
            self.israw = False
            self.ftype = '8bits'
        else:
            self.israw = True
            assert opt.input_nc == 4, "The the input should be 4 channels!!!"
            # self.ftype = '10bits'
            self.ftype = '12bits'
            self.raw_paths = os.path.join(self.rootdir, opt.rawFolder)
            # self.createInputDataFromRaw()

        if opt.warp or opt.mask:
            self.lbridge = CPPbridge('./build/libBridge.so')
            self.w_paths = os.path.join(self.rootdir, opt.wFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
            self.flow_paths = os.path.join(self.rootdir, opt.flowFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
            self.m_paths = os.path.join(self.rootdir, opt.mFolder, opt.nFolder, opt.warp_method, opt.model_input_gen)
            mkdir(self.w_paths)
            mkdir(self.flow_paths)
            mkdir(self.m_paths)

        assert(self.opt.load_size >= self.opt.patch_width)   # patch_width should be smaller than the size of loaded image
        self.input_nc = self.opt.input_nc
        self.output_nc = self.opt.output_nc
        
        gt_paths_list = os.scandir(self.gt_paths)
        if opt.videos is None:
            gt_paths_list = [c.path for c in gt_paths_list if not c.name.startswith('.') and c.is_dir()]
        else:
            opt.videos = opt.videos.split(',')
            gt_paths_list = [c.path for c in gt_paths_list if not c.name.startswith('.') and c.is_dir() and c.name in opt.videos]
        gt_paths_list = sorted(gt_paths_list)
        
        noise_paths_list = os.scandir(self.n_paths)
        if opt.videos is None:
            noise_paths_list = [c.path for c in noise_paths_list if not c.name.startswith('.') and c.is_dir()]
        else:
            noise_paths_list = [c.path for c in noise_paths_list if not c.name.startswith('.') and c.is_dir() and c.name in opt.videos]
        noise_paths_list = sorted(noise_paths_list)

        assert (len(gt_paths_list) == len(noise_paths_list))

        if opt.warp:
            warped_paths_list = [os.path.join(self.w_paths, os.path.basename(d)) for d in gt_paths_list ]
            mkdirs(warped_paths_list)
            

        print ('%d videos' % len(gt_paths_list))
        self.gt_paths_list = gt_paths_list
        self.noise_paths_list = noise_paths_list

        self.patch_depth = self.opt.patch_depth

        self.length = 0
        self.where = []
        self.videos_noisy_path = []
        self.videos_gt_path = []
        self.videos_w_path = []
        self.videos_m_path = []
        self.videos_flow_path = []

        if opt.warp or opt.mask:
            self.createWarpedInputData()

        for v, (video_path, video2_path) in enumerate(zip(self.gt_paths_list, self.noise_paths_list)):
            img_paths = list_video_files_at_dir(video_path)
            img2_paths = list_video_files_at_dir(video2_path)
            assert (len(img_paths) == len(img2_paths))

            self.where = np.concatenate( (self.where, np.array(range(len(img_paths)-self.patch_depth+1),dtype=np.int)+len(self.videos_gt_path) ) )
            self.videos_noisy_path = self.videos_noisy_path + img2_paths
            self.videos_gt_path = self.videos_gt_path + img_paths

            if self.opt.warp:
                for p, path in enumerate(img2_paths):
                    w_path = []
                    m_path = []
                    flow_path = []
                    toCode = os.path.splitext(os.path.basename(path))[0]
                    wfolder = os.path.join(self.w_paths, pathdiff(img2_paths[p], self.n_paths))
                    mfolder = os.path.join(self.m_paths, pathdiff(img2_paths[p], self.n_paths))
                    flowfolder = os.path.join(self.flow_paths, pathdiff(img2_paths[p], self.n_paths))
                    for z in reversed(range(1, self.patch_depth)):
                        fromCode = os.path.splitext(os.path.basename(img2_paths[p-z]))[0]
                        wimagefile = warpedimagefile(wfolder, fromCode, toCode)
                        mimagefile = warpedimagefile(mfolder, fromCode, toCode)
                        flowfile = warpedimagefile(flowfolder, fromCode, toCode)
                        w_path.append(wimagefile)
                        m_path.append(mimagefile)
                        flow_path.append(flowfile)
                    w_path.append(path)
                    self.videos_w_path.append( w_path )
                    self.videos_m_path.append( m_path )
                    self.videos_flow_path.append( flow_path )

        self.where = self.where.astype(int)

    def __len__(self):
        return len(self.where)
    
    def prepare_epoch(self):
        """Load and randomize data"""
        print("nothing to do in prepare_epoch")

    def data_num_channels(self):
        return 3

    def __getitem__(self, index):
        """
        Returns a dictionary that contains gt, n, gt_path, n_path
            gt (tensor) - - Groundtruth images
            n (tensor) - - Noisy images
            gt_path (str) - - Path to the last image in gt
            n_path (str) - - Path to the last image in n
        """

        key = self.where[index]
        gt =  np.asarray([load_image(self.videos_gt_path[key+k],self.ftype) for k in range(self.patch_depth)], dtype=np.float32)
        if self.opt.warp:
            flows = np.asarray([iio_read(path).astype(np.float32) if os.path.isfile(path) else np.zeros(list(gt.shape[1:3])+list([2]), dtype=np.float32)
                                    for path in self.videos_flow_path[key+self.patch_depth-1]], dtype=np.float32)
            flows = flows.transpose(0, 3, 1, 2)
            flows = torch.from_numpy(flows)
        else:
            flows = []


        if self.opt.model_input_gen == 'noisyinputs': 
            noise = np.asarray([load_image(self.videos_noisy_path[key+k],self.ftype)
                                for k in range(self.patch_depth)], dtype=np.float32)
        elif self.opt.model_input_gen == 'lastnoisy':            
            noise = np.asarray([load_image(self.videos_noisy_path[key+self.patch_depth-1],self.ftype)], dtype=np.float32)
            noise = np.concatenate((gt[:-1, :, :, :],  noise), 0)

        gt = gt.transpose(0, 3, 1, 2)
        gt = gt.reshape([gt.shape[0] * gt.shape[1], gt.shape[2], gt.shape[3]])
        gt = gt.transpose(1, 2, 0)

        noise = noise.transpose(0, 3, 1, 2)
        noise = noise.reshape([noise.shape[0] * noise.shape[1], noise.shape[2], noise.shape[3]])
        noise = noise.transpose(1, 2, 0)

        noise = self.T(noise)
        
        if self.opt.warp and self.opt.mask:
            masks = np.array([iio_read(path).astype(np.bool) for path in self.videos_m_path[key+self.patch_depth-1]], dtype=np.float32)
            masks = masks.transpose(0, 3, 1, 2).astype(np.float32)
            masks = torch.from_numpy(masks)

        if hasattr(self.opt, "crop_data") and not self.opt.crop_data is None:
            x, y = [int(s) for s in self.opt.crop_data.split(',')]
            noise = noise[:,:x,:y]
            gt = gt[:x,:y,:]
        return {'gt': self.T(gt), 'n': noise, 'flow': flows, 'gt_path':self.videos_gt_path[key+self.patch_depth-1], 'n_path':self.videos_noisy_path[key+self.patch_depth-1]}
