// Compile with cmake (CMakeLists.txt is provided) or with the following lines in bash:
// g++ -c -fPIC libautosim.cpp -o libautosim.o
// g++ -shared -Wl,-soname,libautosim.so -o libautosim.so libautosim.o


#include <string>
#include <sstream>
#include <iostream>
#include <list>
#include <vector>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>
#include <algorithm>
#include <float.h>
#include <limits>
#include <cmath>
#include <limits>


#ifndef DISABLE_OMP
#include <omp.h>
#endif //DISABLE_OMP

#define PAR_DEFAULT_NPROC 0
#define PAR_DEFAULT_TAU 0.25
#define PAR_DEFAULT_LAMBDA 0.15
#define PAR_DEFAULT_THETA 0.3
#define PAR_DEFAULT_NSCALES 100
#define PAR_DEFAULT_FSCALE 0
#define PAR_DEFAULT_ZFACTOR 0.5
#define PAR_DEFAULT_NWARPS 5
#define PAR_DEFAULT_EPSILON 0.01
#define PAR_DEFAULT_VERBOSE 0

#include "inverse_compositional_algorithm.h"
#include "transformation.h"

// Define C functions for the C++ class - as ctypes can only talk to C...
extern "C"
{

  void inverse_compositional(double *I1, double *I2, int nx, int ny, int nz, double* mat, int nscales, int nparams, int robust, double lambda)
  {
    int verbose, first_scale;
    int delta, nanifoutside, type_gradient;
    double zfactor, TOL;
    // nscales = 0;
    zfactor = 0.5;
    TOL = 0.001;
    // nparams = 8; // PAR_DEFAULT_TYPE
    // robust = 3;
    // lambda = 0.0;
    verbose = 0;
    first_scale = 0;
    delta = 5;
    nanifoutside = 1;
    type_gradient = 3;

    //limit the number of scales according to image size (min 32x32)
    int N = 1 + log(std::min(nx, ny) / 32.) / log(1. / zfactor);
    if (N < nscales || nscales <= 0)
      nscales = N;

    if (verbose)
      printf(
          "\nParameters: scales=%d, zoom=%f, TOL=%f, transform type=%d, "
          "robust function=%d, lambda=%f, delta=%d, "
          "nanifoutside=%d, first scale=%d, gradient type=%d, "
          "\n",
          nscales, zfactor, TOL, nparams, robust, lambda, delta,
          nanifoutside, first_scale, type_gradient);
    
    double *p = new double[nparams];
    pyramidal_inverse_compositional_algorithm(
        I1, I2, p, nparams, nx, ny, nz,
        nscales, zfactor, TOL, robust, lambda, first_scale, nanifoutside,
        delta, type_gradient, verbose);

    params2matrix(p, mat, nparams);
  }

#include "tvl1flow_lib.h"

  void tvl1flow(float *I0, float *I1, float *u, int nx, int ny)
  {

    int fauxval = -1;
    int nproc = (fauxval > 0) ? fauxval : PAR_DEFAULT_NPROC;
    float tau = (fauxval > 0) ? fauxval : PAR_DEFAULT_TAU;
    float lambda = (fauxval > 0) ? fauxval : PAR_DEFAULT_LAMBDA;
    float theta = (fauxval > 0) ? fauxval : PAR_DEFAULT_THETA;
    int nscales = (fauxval > 0) ? fauxval : PAR_DEFAULT_NSCALES;
    int fscale = (fauxval > 0) ? fauxval : PAR_DEFAULT_FSCALE;
    float zfactor = (fauxval > 0) ? fauxval : PAR_DEFAULT_ZFACTOR;
    int nwarps = (fauxval > 0) ? fauxval : PAR_DEFAULT_NWARPS;
    float epsilon = (fauxval > 0) ? fauxval : PAR_DEFAULT_EPSILON;
    int verbose = (fauxval > 0) ? fauxval : PAR_DEFAULT_VERBOSE;
    
    //check parameters
    if (nproc < 0)
    {
      nproc = PAR_DEFAULT_NPROC;
      if (verbose)
        fprintf(stderr, "warning: "
                        "nproc changed to %d\n",
                nproc);
    }
    if (tau <= 0 || tau > 0.25)
    {
      tau = PAR_DEFAULT_TAU;
      if (verbose)
        fprintf(stderr, "warning: "
                        "tau changed to %g\n",
                tau);
    }
    if (lambda <= 0)
    {
      lambda = PAR_DEFAULT_LAMBDA;
      if (verbose)
        fprintf(stderr, "warning: "
                        "lambda changed to %g\n",
                lambda);
    }
    if (theta <= 0)
    {
      theta = PAR_DEFAULT_THETA;
      if (verbose)
        fprintf(stderr, "warning: "
                        "theta changed to %g\n",
                theta);
    }
    if (nscales <= 0)
    {
      nscales = PAR_DEFAULT_NSCALES;
      if (verbose)
        fprintf(stderr, "warning: "
                        "nscales changed to %d\n",
                nscales);
    }
    if (zfactor <= 0 || zfactor >= 1)
    {
      zfactor = PAR_DEFAULT_ZFACTOR;
      if (verbose)
        fprintf(stderr, "warning: "
                        "zfactor changed to %g\n",
                zfactor);
    }
    if (nwarps <= 0)
    {
      nwarps = PAR_DEFAULT_NWARPS;
      if (verbose)
        fprintf(stderr, "warning: "
                        "nwarps changed to %d\n",
                nwarps);
    }
    if (epsilon <= 0)
    {
      epsilon = PAR_DEFAULT_EPSILON;
      if (verbose)
        fprintf(stderr, "warning: "
                        "epsilon changed to %f\n",
                epsilon);
    }

#ifndef DISABLE_OMP
    if (nproc > 0)
      omp_set_num_threads(nproc);
#endif //DISABLE_OMP


    //Set the number of scales according to the size of the
    //images.  The value N is computed to assure that the smaller
    //images of the pyramid don't have a size smaller than 16x16
    const float N = 1 + log(hypot(nx, ny) / 16.0) / log(1 / zfactor);
    if (N < nscales)
      nscales = N;
    if (nscales < fscale)
      fscale = nscales;

    if (verbose)
      fprintf(stderr,
              "nproc=%d tau=%f lambda=%f theta=%f nscales=%d "
              "zfactor=%f nwarps=%d epsilon=%g\n",
              nproc, tau, lambda, theta, nscales,
              zfactor, nwarps, epsilon);

    // ----- u is allocated by python
    // //allocate memory for the flow
    // u = xmalloc((size_t)(2 * nx * ny * sizeof *u));
    float *v = u + nx * ny;

    //compute the optical flow
    Dual_TVL1_optic_flow_multiscale(
        I0, I1, u, v, nx, ny, tau, lambda, theta,
        nscales, fscale, zfactor, nwarps, epsilon, verbose);

    
    // ---- The memory will be controlled by python ----
    // //delete allocated memory
    // free(I0);
    // free(I1);
    // free(u);
  }

  #include "downsa.h"
  void downsa(float *x, float *y, int w, int h, int pd, int n, char type)
  {
  	downsa2d(y, x, w, h, pd, n, type);
  }

  #include "upsa.h"
  void upsa(float *x, float *y, int w, int h, int pd, int n, int zt, float off_x, float off_y)
  {
    int W = n * w - n; // l'amour est enfant de Bohême
    int H = n * h - n; // il n'a jamais jamais connu de loi
    float nf = n;
    for (int j = 0; j < H; j++)
      for (int i = 0; i < W; i++)
      {
        float tmp[pd];
        interpolate_vec(tmp, x, w, h, pd, (i - off_x) / nf, (j - off_y) / nf, zt);
        for (int l = 0; l < pd; l++)
          setsample(y, W, H, pd, i, j, l, tmp[l]);
      }
  }

  #include "blur.h"
  #include "parsenumbers.h"
    void blur(float *x, float *y, int w, int h, int pd, char *kernel_id, char *kernel_params, bool boundary_symmetric, bool boundary_periodic, bool boundary_zero) //char* opts)
    {
#ifdef _USEFFTW
      int maxparam = 10;
      float param[maxparam];
      int nparams = parse_floats(param, maxparam, kernel_params);
      if (nparams < 1)
        fail("please, give at least one parameter");

      if (boundary_symmetric || boundary_zero || boundary_periodic)
      {
        int ww = 2 * w, hh = 2 * h;
        float *xx = (float *) xmalloc(ww * hh * pd * sizeof *xx);
        float *yy = (float *) xmalloc(ww * hh * pd * sizeof *xx);
        for (int j = 0; j < h; j++)
          for (int i = 0; i < w; i++)
            for (int l = 0; l < pd; l++)
            {
              float g = x[(j * w + i) * pd + l];
              float g0 = boundary_zero ? 0 : g;
              int si = boundary_periodic ? w + i : 2 * w - i - 1;
              int sj = boundary_periodic ? h + j : 2 * h - j - 1;
              xx[(j * ww + i) * pd + l] = g;
              xx[(sj * ww + i) * pd + l] = g0;
              xx[(j * ww + si) * pd + l] = g0;
              xx[(sj * ww + si) * pd + l] = g0;
            }
        blur_2d(yy, xx, ww, hh, pd, kernel_id, param, nparams);
        for (int j = 0; j < h; j++)
          for (int i = 0; i < w; i++)
            for (int l = 0; l < pd; l++)
              y[(j * w + i) * pd + l] = yy[(j * ww + i) * pd + l];
        free(xx);
        free(yy);
      }
      else
        blur_2d(y, x, w, h, pd, kernel_id, param, nparams);
#else
      fail("The function blur (in libBridge.cpp) depends on the FFTW external library. Please activate its flag in cmake file !!!");
#endif
    }


  }