from copy import deepcopy
import torch
from .base_model import BaseModel
import networks
import kornia
from util.flow_utils import warp
import random

class recurrentFeatModel(BaseModel):
    """ This class implements a "recurrent" model for denoising a sequence 
        of n-1 previously denoised frames plus the current noisy frame 
        from a video.
    """
    @staticmethod
    def modify_commandline_options(parser, is_train=True):
        """Add new specific options, and rewrite default values for existing options.

        Parameters:
            parser          -- original option parser
            is_train (bool) -- whether training phase or test phase. You can
                               use this flag to add training-specific or
                               test-specific options.

        Returns:
            the modified parser.
        """
        BaseModel.modify_commandline_options(parser, is_train)
        if is_train:
            parser.set_defaults(patch_depth=3, no_val=False,patch_width=96, val_dataset_mode='infer4rec')
            #parser.add_argument('--band_width', type=int, default=0, help='Sets the band width for each patch in training. On this band the loss is not computed. It should be reflected in the ground truth image. ')
            # parser.add_argument('--warpunknown_from_n', action='store_true', default=False, help='Use current noisy frame to complete out of patch information when warping.')
            # parser.add_argument('--input_detach', action='store_true', default=False, help='Detach all inputs i.e. we do not backprop beyond the current unroll.')
        parser.add_argument('--lambda_L1', type=float, default=100.0, help='weight for L1 loss')
        parser.add_argument('--model_patch_depth', type=int, default=2, help='True model patch_depth (should be < patch_depth). Corresponds to the ancient basic model patch_depth.')
        parser.add_argument('--unroll_focus', type=str, default="all", help='Types of losses for recurrency: [all, ge_j, last, incr_j, gradualj, gradualjj_fromjj, gradunijj, gradunijj_fromjj], where j>0 and jj must have two digits')
        # parser.add_argument('--residual_rec', default=False, action='store_true', help='Add a skip connection from the previous frame (after warping) to the network output.')
        parser.add_argument('--no_frame_rec', default=False, action='store_true', help='Do not use frame recurrence.')
        parser.add_argument('--freeze_transform', default=False, action='store_true', help='Set freq and color transforms untrainable.')
        parser.add_argument('--ISO', type=int, default=3200, help='The ISO of dataset, 3200 or 12800.')
        parser.add_argument('--lambda_pre_L1', type=float, default=0, help='weight for L1 loss for predenoised frame')
        return parser

    
    def set_visuals(self, visuals):
        # specify the images you want to save/display. The training/test
        # scripts will call <BaseModel.get_current_visuals>
        if visuals == "denoiser":
            self.visual_names = ['denoised']
        elif visuals == "debug":
            self.visual_names = ['n_1', 'n_2', 'gt_1', 'gt_2', 'denoised']
        elif visuals == "none":
            self.visual_names = []
        elif visuals == "fusion":
            self.visual_names = ['fusion', 'denoised', 'weights', 'n_map', 'f_map', 'refine_w','denoise_net','temporal_aver']
        elif visuals == "weights":
            self.visual_names = ['denoised', 'refine_w', 'n_map', 'f_map','occ_mask']
        

    def __init__(self, opt):
        """Initialize the recurrent class.

        Parameters:
            opt (Option class)-- stores all the experiment flags; needs to be a
                                 subclass of BaseOptions
        """
        BaseModel.__init__(self, opt)

        # define loss function (this is the one that will be used during training)
        self.criterionL1 = torch.nn.L1Loss()
        self.tv = kornia.losses.TotalVariation()

        # specify other losses to print during training/validation/testing. The training/test
        # scripts will call <BaseModel.get_current_losses>
        self.loss_names = ['L1', 'PSNR', 'Denoiser', 'freq', 'color', 'predenoise_L1']

        # specify the images you want to save/display. The training/test
        # scripts will call <BaseModel.get_current_visuals>
        self.set_visuals(opt.visuals)

        # specify the models you want to save to the disk. The training/test
        # scripts will call <BaseModel.save_networks> and
        # <BaseModel.load_networks>
        self.model_names = ['Denoise']

        # define networks
        if opt.no_frame_rec:
            network_input_nc = opt.input_nc

        self.netDenoise = networks.define_net_arch(network_input_nc, opt.output_nc, opt.ngf,
                                                   opt.netDenoiser, opt.norm, opt.Nmomentum,
                                                   not opt.no_dropout, opt.init_type,
                                                   opt.init_gain, self.gpu_ids, NoPF=self.opt.model_patch_depth-1,freezeTrans=opt.freeze_transform)        
        if isinstance(self.netDenoise, torch.nn.DataParallel):
            self._netDenoise = self.netDenoise.module
        else:
            self._netDenoise = self.netDenoise

        if self.isTrain:
            # initialize optimizers; schedulers will be automatically created
            # by function <BaseModel.setup>.
            self.optimizer_Denoise = torch.optim.Adam(self.netDenoise.parameters(), lr=opt.lr,
                                                      betas=(opt.beta1, 0.999))
            self.optimizers.append(self.optimizer_Denoise)

            # remove band around the image border - those pixels will be computed
            # by the denoising network, but they will be multiplied by zero in the loss
            # if self.opt.band_width>0:
            #     self.active_loss = torch.zeros((opt.batch_size, opt.output_nc,
            #                                     opt.patch_width, opt.patch_width),
            #                                    device=self.device, dtype=torch.float32)
            #     self.active_loss[:,:,self.opt.band_width:-self.opt.band_width,self.opt.band_width:-self.opt.band_width] = 1

    def set_input(self, input):
        """Unpack input data from the dataloader and perform necessary pre-processing steps.

        Parameters:
            input (dict): include the data itself and its metadata information.
        """

        self.image_paths = input['n_path']
        self.gt = input['gt'].to(self.device, non_blocking=self.opt.non_blocking)
        self.n  = input['n' ].to(self.device, non_blocking=self.opt.non_blocking)

        if not self.isTrain and input['FirstOfVideo']:
            D  = self.opt.model_patch_depth - 1
            C  = self.opt.input_nc
            if not self.opt.no_frame_rec:
                if self.opt.model_input_gen == "lastnoisy":
                    # initialize recurrence with previous ground truth frame (fake recurrency)
                    self.lastden = input['gt'][:, 0:D*C,:,:].to(self.device, non_blocking=self.opt.non_blocking)
                else:
                    # initialize recurrence with previous noisy frame
                    self.lastden = input['n'][:, 0:D*C,:,:].to(self.device, non_blocking=self.opt.non_blocking)
            B, _, H, W = self.n.size()
            # self.lastfeat = self._netDenoise.get_rec_nil_features(B,H,W, device=self.device, non_blocking=self.opt.non_blocking)
            self.lastfeat = self._netDenoise.get_rec_init_features(self.n[:,0:C,:,:])

        # if not self.isTrain:
        #     self.lastden_from_script = input['denoised'].to(self.device, non_blocking=self.opt.non_blocking)
        # print(torch.all(self.lastden_from_script==self.lastden))

        if self.opt.warp:
            self.flow = input['flow'].to(self.device, non_blocking=self.opt.non_blocking)

            if self.opt.mask:
                self.mask = input['mask'].to(self.device, non_blocking=self.opt.non_blocking)

        # add unrollings dimension in testing for 
        # compatibility with training
        if not self.isTrain and self.opt.warp:
            self.flow = self.flow.unsqueeze_(1)
            if self.opt.mask:
                self.mask = self.mask.unsqueeze_(1)

        if self.isTrain:
            # information about the training time
            self.epoch = input['epoch']
            self.epoch_iter = input['epoch_iter']
            self.epoch_length = input['epoch_length']


    def forward(self):
        """Run forward pass; called by both functions <optimize_parameters> and <test>."""
        B, _, H, W = self.n.size()
        TD = self.opt.patch_depth - 1 if self.isTrain else 1
        D  = self.opt.model_patch_depth - 1
        C  = self.opt.input_nc
        # B   : batch size
        # TD  : maximum number of unrollings
        # D   : number of previous frames given to the network
        # C   : number of channels of the input images
        # H,W : height and width of the frame

        ###########################################################################
        # NOTE
        # This code considers that the network can take more than one previous frame.
        # D is the number of previous frames given to the network as input.
        # That is, for frame t the denoised frame is computed as follows:
        #
        # denoised[t] = net(denoised[t-D] warped to t,
        #                   ...
        #                   denoised[t-2] warped to t,
        #                   denoised[t-1] warped to t,
        #                   noisy[t])
        #
        # This can be controlled with the command line argument --model_patch_depth
        # (model_patch_depth is the total number of input frames including
        # the current noisy frame, i.e. D+1).
        #
        # Until now we have only used '--model_patch_depth 2' (the current
        # noisy frame and only the previous denoised frame) and therefore D=1.
        #
        #
        # During training the dataloader extract from the videos 3D crops of size
        #
        # patch_width x patch_width x patch_depth
        #
        # patch_depth is the number of frames (or cropped frames) we have
        # available, and it needs to be larger or equal than model_patch_depth.
        #
        # Suppose for example than model_patch_depth = 5, and patch_depth = 10.
        # In that case we can do 6 unrollings (patch_depth - model_patch_depth + 1):
        #
        # Notation: n0 n1 n2 n3 n4 n5 n6 n7 n8 n9 are the noisy frames
        #
        # d4 = net(n0-warped-to-4, n1-warped-to-4, n2-warped-to-4, n3-warped-to-4, n4)
        # d5 = net(n1-warped-to-5, n2-warped-to-5, n3-warped-to-5, d4-warped-to-5, n5)
        # d6 = net(n2-warped-to-6, n3-warped-to-6, d4-warped-to-6, d5-warped-to-6, n6)
        # d7 = net(n3-warped-to-7, d4-warped-to-7, d5-warped-to-7, d6-warped-to-7, n7)
        # d8 = net(d4-warped-to-8, d5-warped-to-8, d6-warped-to-8, d7-warped-to-8, n8)
        # d9 = net(d5-warped-to-9, d6-warped-to-9, d7-warped-to-9, d8-warped-to-9, n9)
        #
        # Once a frame is denoised, it is used in the next unrollings. Else, the noisy
        # frames are used. In order to compute the necessary warpings, we need these
        # optical flows:
        #
        # flow 4 -> 0, flow 4 -> 1, flow 4 -> 2, flow 4 -> 3
        # flow 5 -> 1, flow 5 -> 2, flow 5 -> 3, flow 5 -> 4
        # flow 6 -> 2, flow 6 -> 3, flow 6 -> 4, flow 6 -> 5
        # flow 7 -> 3, flow 7 -> 4, flow 7 -> 5, flow 7 -> 6
        # flow 8 -> 4, flow 8 -> 5, flow 8 -> 6, flow 8 -> 7
        # flow 9 -> 5, flow 9 -> 6, flow 9 -> 7, flow 9 -> 8
        #
        # Thus we need D flows for each unrolling. Let's denote the numer
        # of unrollings by TD = patch_depth - model_patch_depth + 1.
        #
        # The dataloader stores all these flows for a batch of size B in a
        # tensor of dimensions  B x TD x D x 2 x W x H.
        #
        ###########################################################################

        unrollings = TD
        self._netDenoise.set_transforms()
        if self.isTrain:
            # initialize with the first noisy frame
            self.lastfeat = self._netDenoise.get_rec_init_features(self.n[:,0:C,:,:])

            # determine the number of unrollings (the number of unrollings can change
            # during training; e.g. the gradualii_fromjj trains with one unrolling until
            # epoch jj is reached. This is mainly to speed up the training of recurrent
            # networks)
            if self.opt.unroll_focus[:5] == "gradu" and self.opt.unroll_focus[-7:-2] == "_from":

                # epoch where the recurrent training starts
                epoch_start_rec = float(self.opt.unroll_focus[-2:])

                # number of unrollings
                unrollings = 1 if self.epoch < epoch_start_rec else TD

            # during training, we store the output of the unrollings here
            self.denoised_list = []
            # during training, we store the fusion weight (multiplied to temporal averaging), so to add TV regularization to them
            self.fusionW_list = []
            self.predenoised_list = []

        for a in range(unrollings): # unrolling

            # collect the previous frames for the network input
            # NOTE: so far we have always used D=1, so here we only colect
            #       the previous warped frame
            netinput = None
            featinput = []
            for f in self.lastfeat:
                if f is None:
                    featinput.append(None)
                else:
                    featinput.append( f )

            for b in range(D):  # for each previous frame

                if self.opt.warp:   # warp the previous feature to current noisy frame
                    for i,f in enumerate(featinput):
                        if not f is None:
                            Bf,Cf,Hf,Wf = f.size()
                            onefC = int(Cf / (self._netDenoise.NoPF))
                            featinput[i][:,b*onefC:(b+1)*onefC,:,:] = warp(featinput[i][:,b*onefC:(b+1)*onefC,:,:].clone(), self.flow[:, a, b, :, :, :], interp="bicubic")[0]
                                
                # else:
                #     # case of no warping: nothing to do here
                #     warped = img1

            # add current noisy patch to the input
            if not self.opt.no_frame_rec:
                netinput = torch.cat((netinput, self.n[:,(a+D)*C:(a+D+1)*C,:,:]), 1)
            else:
                netinput = self.n[:,(a+D)*C:(a+D+1)*C,:,:]
            
            # set warped feature for use (downsampling)
            self._netDenoise.set_rec_features(featinput)

            # denoise frame
            # if self.isTrain and self.opt.input_detach:
            #     # this option detaches the unrollings - the gradient will not bachprop
            #     # through the unrollings
            #     self.denoised = self.netDenoise(netinput.detach())
            # else:
                # this is the standard recurrent training with backprop through the unrollings
            self.denoised = self.netDenoise(netinput,self.opt.ratioVariance)   

            self.fusionW = self._netDenoise.fusion_weights

            self.predenoised = self._netDenoise.get_denoised()

            # if self.opt.residual_rec:
            #     # skip connection from previous warped input to output
            #     self.denoised = self.denoised + netinput[:,:C,:,:]

            if self.isTrain:
                # add denoised frame to list
                self.denoised_list.append( self.denoised ) 
                # add fusion weight to the list
                self.fusionW_list.append( self.fusionW )
                self.predenoised_list.append( self.predenoised )

            # update the newest recurrence with current temporal fusion
            tfeat = self._netDenoise.get_current_features()
            for i,f in enumerate(self.lastfeat):
                if not f is None:
                    Bf,Cf,Hf,Wf = tfeat[i].size()
                    self.lastfeat[i] = torch.cat( (f[:,Cf:,:,:],  tfeat[i]), 1)

        if self.isTrain:
            # store input of last unrolling (used only for visualization)
            self.netinput = netinput


    def compute_visuals(self):
        """Calculate additional output images for visdom and visualization"""
        C  = self.opt.input_nc
        i = random.randint(0, self.opt.batch_size-1)

        if not self.isTrain:
            i = 0

        if self.isTrain:
            B, _, H, W = self.n.size()
            TD = self.opt.patch_depth - 1
            D  = self.opt.model_patch_depth - 1

            a = TD-1
            self.gt_1 = self.gt[i:(i+1), a*C:(a+1)*C,:,:] # first groundtruth patch
            self.gt_2 = self.gt[i:(i+1),(a+D)*C:(a+1+D)*C,:,:] # last groundtruth patch
            if not self.opt.no_frame_rec:
                if self.opt.mask and self.opt.mask_on_chan:
                    self.n_1 = self.netinput[i:(i+1),  0:C,:, :]*self.netinput[i:(i+1), C:(C+1),:,:] \
                            + -1*(1.0-self.netinput[i:(i+1), C:(C+1),:,:])
                    self.n_2 = self.netinput[i:(i+1), -C:, :, :]
                else:
                    self.n_1 = self.netinput[i:(i+1), 0:C,:,:]
                    self.n_2 = self.netinput[i:(i+1), D*C:(D+1)*C,:,:]
            else:
                self.n_1 = torch.zeros_like(self.netinput[i:(i+1), :,:,:])
                self.n_2 = self.netinput[i:(i+1), :,:,:]
        else:

            self.gt_1 = self.gt[i:(i+1), -2*C:-C,:,:] # first groundtruth patch
            self.gt_2 = self.gt[i:(i+1),-C: ,:,:] # last groundtruth patch

            if not self.opt.no_frame_rec:
                self.n_1 = torch.zeros_like(self.n[i:(i+1),-C:,:,:]) # lastden is now equal to denoised, so it's of no use in here
                self.n_2 = self.n[i:(i+1),-C:,:,:] # last noisy patch
            else:
                self.n_1 = torch.zeros_like(self.n[i:(i+1),-C:,:,:])
                self.n_2 = self.n[i:(i+1),-C:,:,:]

            if self.opt.warp and False:
                # warp frame using flow - also returns mask of invalid warped pixels
                # these are pixels that fell outside of the previous frame
                self.n_1, w_mask = warp(self.n_1, self.flow[:, 0, 0, :, :, :], interp="bicubic")

                if self.opt.mask:
                    # mask is the union of the invalid warping mask with the
                    # warping error mask (which is precomputed using
                    # the noisy frames and provided by the dataloader)
                    mask = w_mask*self.mask[:, 0, 0, :, :, :]

                    if self.opt.mask_on_chan:
                        # mask as an additional input channel
                        self.n_1 = torch.cat((self.n_1,mask), 1)
                    else:
                        # mask encoded in the warped frame by setting the masked pixels as mask_value
                        self.n_1 = self.opt.mask_value*(1.0-mask) + mask*self.n_1

        # if self.opt.val_visuals == "fusion":
        if "fusion" in self.visual_names:
            self.fusion = self._netDenoise.get_fused()
            self.weights = self._netDenoise.get_fusion_weights()
            self.n_map = self._netDenoise.get_noisy_noisemap()
            self.f_map = self._netDenoise.get_fusion_noisemap()
            self.refine_w = self._netDenoise.get_refine_weights() 
            self.denoise_net = self._netDenoise.get_denoised()
            self.temporal_aver = self._netDenoise.get_previous_ta()
        if "weights" in self.visual_names:
            self.weights = self._netDenoise.get_fusion_weights()
            self.refine_w = self._netDenoise.get_refine_weights() 
            # self.occ_mask = self._netDenoise.get_occmask()
            self.n_map = self._netDenoise.get_noisy_noisemap()
            self.f_map = self._netDenoise.get_fusion_noisemap()



    def compute_losses(self):
        """Compute losses; called in every training and validation iteration"""
        """Total loss is composed of three parts, including loss on the frames, loss on the orthogonality color and frequency transform"""  
        color_mul = torch.matmul(self._netDenoise.ct.filter_color.weight.squeeze(), \
                                    self._netDenoise.ct.inv_filter_color.weight.squeeze())
        self.loss_color = ((color_mul - torch.eye(color_mul.shape[0], device=self.device))**2).sum()

        freq_mul = torch.matmul(self._netDenoise.ft.dec_filter1d, self._netDenoise.ft.rec_filter1d.transpose(0,1) )
        self.loss_freq = ((freq_mul - torch.eye(freq_mul.shape[0], device=self.device))**2).sum()
        
        if self.isTrain:
            B, _, H, W = self.n.size()
            TD = self.opt.patch_depth - 1
            D  = self.opt.model_patch_depth - 1
            # B   : batch size
            # TD  : maximum number of unrollings
            # D   : number of frames given to the network (we never used more than 2)
            # H,W : height and width of the frame

            # During training, the loss is a weighted sum of the individual losses
            # of the denoised frames for each unrolling. The weights are controlled
            # via the command line flag --unroll_focus.
            #
            # all:     uniform weights                                [1/6, 1/6, 1/6, 1/6, 1/6, 1/6]
            # last:    only last unrolling                            [ 0 ,  0 ,  0 ,  0 ,  0 ,  1 ]
            # ge[ii]:  only unrolling greater or equal than ii. ge3:  [ 0 ,  0 ,  0 , 1/3, 1/3, 1/3]
            # inc[ee]: weight for unrolling u is prop to u**ee  inc2: [ 1 ,  4 ,  9 , 16 , 25 , 36 ]/91
            # graduni[ii]_from[jj]: gradual transition from non-recurrent to uniform,
            #                       between epochs jj and ii+jj
            # gradual[ii]_from[jj]: gradual transition from non-recurrent to 90% of weight
            #                       in the last unrolling, between epochs jj and ii+jj
            #
            # See below for more details about the gradual transitions.

            # determine the number of unrollings (read comments in the forward() function)
            unrollings = TD
            if self.opt.unroll_focus[:5] == "gradu" and self.opt.unroll_focus[-7:-2] == "_from":
                epoch_start_rec = int(self.opt.unroll_focus[-2:])
                unrollings = 1 if self.epoch < epoch_start_rec else TD


            # compute individual losses for the denoised frames produced by the unrollings
            # loss_L1_vec, loss_PSNR_vec = [], []
            loss_L1_vec, loss_PSNR_vec, loss_predenoised_vec = [], [], []
            for a in range(unrollings):
                # the following flag determines if the loss associated to the output
                # of unrolling a should be computed and added to the total loss
                add_flag = self.opt.unroll_focus == "all" \
                        or (self.opt.unroll_focus == "last" and a == TD-1) \
                        or (self.opt.unroll_focus[:2] == "ge" and a >= int(self.opt.unroll_focus[3:])) \
                        or self.opt.unroll_focus[:4] == "incr" \
                        or self.opt.unroll_focus[:5] == "gradu"

                if add_flag:
                    gt_2 = self.gt[:,(a+D)*self.opt.input_nc:(a+1+D)*self.opt.input_nc,:,:]
                    loss_L1_vec.append( self.criterionL1(self.denoised_list[a], gt_2) * self.opt.lambda_L1 )
                    loss_PSNR_vec.append( kornia.losses.psnr_loss(self.denoised_list[a], gt_2, 2.0) )
                    loss_predenoised_vec.append( self.criterionL1( self.predenoised_list[a], gt_2 ) * self.opt.lambda_pre_L1 )

            # compute weights for aggregating the individual losses into the total loss
            if self.opt.unroll_focus[:4] == "incr":
                # 'incr[e]' determines the weights as [i+1]**e, where i is the unrolling index.
                # Example for 4 unrollings and 'incr2' we will have weights = [1, 2**2, 3**2, 4**2]
                # (these weights are then normalized to sum 1)
                weigths = torch.arange(1, TD+1, device=self.device) ** float(self.opt.unroll_focus[5:])
                weigths = weigths/torch.sum(weigths)
                self.loss_L1   = torch.sum( weigths * torch.stack(loss_L1_vec, 0) )
                self.loss_PSNR = torch.sum( weigths * torch.stack(loss_PSNR_vec, 0) )

            elif self.opt.unroll_focus[:5] == "gradu":
                # These are dynamic weights that change during training. The
                # goal is to have a gradual transition between the non-recurrent
                # and the recurrent training.
                #
                # The non-recurrent training is obtained by focusing only in the
                # first unrolling:
                # weights0 = [1, 0, 0, ...., 0] (all weights in the first unrolling)
                #
                # Depending on the final weights of the transition, we have two
                # options 'gradual[ii]_from[jj]' and 'graduni[ii]_from[jj]':
                #
                # - graduni: uniform weights 1/TD*[1, 1, ..., 1]
                #
                # - gradual: 90% of the final weights on the last unrolling
                #            the remaining 10% is split uniformly on the
                #            rest of the unrollings
                #
                #            [1/10/(TD-1), 1/10/(TD-1), ..., 9/10]
                #
                # The transition is controled by ii and jj. It starts at the
                # beginning of epoch jj, and ends at the end of epoch jj+ii.
                # That is, it takes ii+1 epochs. The '_from[jj]' suffix is optional,
                # in that case jj is 1 and the transtion starts at the first epoch.
                #
                # The transition is linear in the case of graduni[ii]_from[jj]. In
                # the case of gradual[ii]_from[jj], it is divided into two linear
                # segments of equal length: first a transition to uniform weights
                # and then from uniform weights to the 90% on the last unrolling.
                #
                # To make the transition very smooth, the weights are updated
                # on every **training iteration**.
                #
                # Note that while the jj-th epoch is not reached, the weights are
                # zero except for the first unrolling. Thus, to speed up training
                # we only compute the first unrolling for epochs < jj.



                # epoch1 is where the gradual change in the weights starts (jj)
                if self.opt.unroll_focus[-7:-2] == "_from":
                    epoch1 = int(self.opt.unroll_focus[-2:])
                else:
                    epoch1 = 1

                # epoch2 is where the gradual change in the weights stops (jj+ii)
                epoch2 = float(self.opt.unroll_focus[7:9]) + epoch1

                if self.epoch < epoch1:
                    # in this case, a single unrolling is computed. weights is a number
                    weights = 1.

                else:
                    # initial weights in the transition
                    weights0 = torch.zeros(TD, device=self.device)
                    weights0[0] = 1.

                    # weights from the epoch2 until end of training
                    if self.opt.unroll_focus[4:7] == 'uni':
                        # with 'graduni' final weights are uniform
                        weights2 = 1./float(TD) * torch.ones(TD, device=self.device)

                        # and intermediante weights (at (epoch2+epoch1)/2 are just
                        # the average of the initial and final weights
                        weights1 = .5*(weights0 + weights2)

                    else:
                        # with 'gradual' final weights put 90% of the mass in
                        # the last unrolling
                        weights2 = .1/float(TD-1)*torch.ones(TD, device=self.device)
                        weights2[TD-1] = .9

                        # and intermediante weights (at (epoch2+epoch1)/2 are uniform
                        weights1 = 1./float(TD) * torch.ones(TD, device=self.device)

                    if self.epoch >= epoch2:
                        # use the final weights
                        weights = weights2

                    else:
                        # progress (but between 0 and 2 instead of 0 and 1)
                        tr_progress = 2. * min(1., 1./(epoch2 - epoch1) * (self.epoch - epoch1 + \
                                               float(self.epoch_iter)/float(self.epoch_length)))

                        if tr_progress < 1.:
                            # between 0 and epoch2/2: transition between weights0 and weights1
                            weights = (1. - tr_progress) * weights0 + tr_progress * weights1

                        else:
                            # between epoch2/2 and epoch2: transition between weights1 and weights2
                            tr_progress = tr_progress - 1.
                            weights = (1. - tr_progress) * weights1 + tr_progress * weights2

                #print('epoch %d - iter %d : weights' % (self.epoch, self.epoch_iter))
                #print(weights)

                self.loss_L1   = torch.sum( weights * torch.stack(loss_L1_vec  , 0) )
                self.loss_PSNR = torch.sum( weights * torch.stack(loss_PSNR_vec, 0) )
                self.loss_predenoise_L1 = torch.sum( weights * torch.stack(loss_predenoised_vec, 0) )

            else:
                # uniform weighting of the individual frame losses
                self.loss_L1   = torch.mean(torch.stack(loss_L1_vec, 0))
                self.loss_PSNR = torch.mean(torch.stack(loss_PSNR_vec, 0))
                self.loss_predenoise_L1 = torch.mean( torch.stack(loss_predenoised_vec, 0) )

            # store loss to be optimized
            self.loss_Denoiser =  self.loss_L1 +self.loss_color+ self.loss_freq + self.loss_predenoise_L1
            

        else:
            # testing: only compute loss for current frame
            gt_2 = self.gt[:,-self.opt.input_nc:,:,:] # last groundtruth patch
            self.loss_L1 = self.criterionL1(self.denoised, gt_2) * self.opt.lambda_L1
            self.loss_PSNR = kornia.losses.psnr_loss(self.denoised, gt_2, 2.0)

            self.loss_predenoise_L1 = self.criterionL1( self.predenoised,gt_2)
            self.loss_Denoiser = self.loss_L1 +self.loss_color+ self.loss_freq + self.loss_predenoise_L1

    def backward_Denoiser(self):
        """Compute loss and calculate gradients for netDenoise"""
        self.compute_losses()
        self.loss_Denoiser.backward()

    def optimize_parameters(self):
        """All steps needed to optimize netDenoise in each iteration"""
        self.forward()                      # apply denoiser
        self.optimizer_Denoise.zero_grad()  # set gradients to zero
        self.backward_Denoiser()            # calculate gradients
        self.optimizer_Denoise.step()       # update weights
