from statistics import variance
import torch
import torch.nn as nn
import numpy as np
import torch.nn.functional as F


device = 'cuda' if torch.cuda.is_available() else 'cpu'

class FCNN(nn.Module):
    '''
    Fusion stage, predicting weights for fusion, which is a convex
    combination of two frames.
    '''

    def __init__(self, in_channels, features=16, out_channels=1, num_layers=2):
        super().__init__()

        kernel_size = 3
        layers = []

        # first layer
        layers.append(
            nn.Conv2d(in_channels, features, kernel_size, padding=1)
        )
        layers.append( nn.ReLU()  )
        # mid layers
        for _ in range(num_layers-2):
            layers.append(
                nn.Conv2d(features, features, kernel_size,padding=1)
            )
            layers.append(nn.ReLU())

        # last layer
        layers.append(
            nn.Conv2d(features, out_channels, kernel_size, padding=1)
        )
        layers.append(nn.Sigmoid())

        self.fcnn = nn.Sequential(*layers)

    def forward(self, x):
        # Input: noisy LL subband, fused LL subbang, and (uniform) noise map
        # Output: two tensors of weight at each pixel position.
        w1 = self.fcnn(x)
        w2 = 1 - w1
        return w1, w2


class DCNN(nn.Module):
    '''
    Denoising stage, output denoised frame.
    '''
    def __init__(self, in_channels, features=16, out_channels=16, num_layers=2):
        super().__init__()

        kernel_size = 3
        layers = []

        # first layer
        layers.append(
            nn.Conv2d(in_channels, features, kernel_size, padding=1)
        )
        layers.append( nn.ReLU() )
        # mid layers
        for _ in range(num_layers-2):
            layers.append(
                nn.Conv2d(features, features, kernel_size,padding=1)
            )
            layers.append( nn.ReLU() )

        # last layer
        layers.append(
            nn.Conv2d(features, out_channels, kernel_size, padding=1)
        )

        self.dcnn = nn.Sequential(*layers)


    def forward(self, x):
        # Input: Fused frame (all subbands), noisy LL subband, and noise map
        # Output: Denoised frame (all subbands)
        out = self.dcnn(x)

        return out

class RCNN(nn.Module):
    '''
    Refinement stage, predicting weights for refinement, which is also a convex
    combination of two frames.
    '''
    def __init__(self, in_channels, features=16, out_channels=1, num_layers=2):
        super().__init__()

        kernel_size = 3
        layers = []

        # first layer
        layers.append(
            nn.Conv2d(in_channels, features, kernel_size, padding=1)
        )
        layers.append( nn.ReLU() )
        # mid layers
        for _ in range(num_layers-2):
            layers.append(
                nn.Conv2d(features, features, kernel_size,padding=1)
            )
            layers.append( nn.ReLU() )

        # last layer
        layers.append(
            nn.Conv2d(features, out_channels, kernel_size, padding=1)
        )
        layers.append(nn.Sigmoid())

        self.rcnn = nn.Sequential(*layers)


    def forward(self, x):
        # Input: Denoised & fused frame (all subbands), noise map
        # alpha decides the maximum mass we put on the fusion
        # Output: Two tensors of weights at each pixel position.
        w1 = self.rcnn(x)
        w2 = 1 - w1
        return w1, w2

class frequency_transform(nn.Module):
    '''Implemente frequency transform initialied by haar transform'''
    def __init__(self, in_channels, init_wavelet = 'haar'):
        super().__init__()
        import pywt 
        coe_wavelet = pywt.Wavelet(init_wavelet)
        dec_filter1d = np.stack([coe_wavelet.dec_lo[::-1], coe_wavelet.dec_hi[::-1]],axis=0 )
        rec_filter1d = np.stack([coe_wavelet.rec_lo, coe_wavelet.rec_hi], axis=0)

        device = 'cuda' if torch.cuda.is_available() else 'cpu'

        self.dec_filter1d = nn.Parameter(torch.tensor(dec_filter1d,dtype=torch.float, device=device), requires_grad=False )
        self.rec_filter1d = nn.Parameter(torch.tensor(rec_filter1d,dtype=torch.float, device=device), requires_grad=False )

        # print(self.dec_filter1d)   # 2-by-2 tensor
        # print(self.rec_filter1d)   # 2-by-2 tensor
        print(torch.matmul(self.dec_filter1d,self.rec_filter1d.transpose(1,0) ))

        self.in_channels = in_channels

        self.dec_filter2d_()
        self.rec_filter2d_()

    def dec_filter2d_(self):
        # 2d frequency transform 

        dec_LL = torch.matmul(self.dec_filter1d[:1].transpose(0,1), self.dec_filter1d[:1] ) 
        dec_LH = torch.matmul(self.dec_filter1d[:1].transpose(0,1), self.dec_filter1d[1:2] ) 
        dec_HL = torch.matmul(self.dec_filter1d[1:2].transpose(0,1), self.dec_filter1d[:1] ) 
        dec_HH = torch.matmul(self.dec_filter1d[1:2].transpose(0,1), self.dec_filter1d[1:2] ) 
        
        dec_2d = torch.stack([dec_LL,dec_LH,dec_HL,dec_HH],dim=0)

        self.dec_filter2d = dec_2d.repeat(self.in_channels,1,1).unsqueeze(1)
        # return self.dec_filter1d


    def rec_filter2d_(self):        
        rec_LL = torch.matmul(self.rec_filter1d[:1].transpose(0,1),self.rec_filter1d[:1] ) 
        rec_LH = torch.matmul(self.rec_filter1d[:1].transpose(0,1),self.rec_filter1d[1:2] ) 
        rec_HL = torch.matmul(self.rec_filter1d[1:2].transpose(0,1),self.rec_filter1d[:1] ) 
        rec_HH = torch.matmul(self.rec_filter1d[1:2].transpose(0,1),self.rec_filter1d[1:2] ) 
        
        rec_2d = torch.stack([rec_LL,rec_LH,rec_HL,rec_HH],dim=0)

        self.rec_filter2d = rec_2d.repeat(self.in_channels,1,1).unsqueeze(1)
        # return self.rec_filter1d

class color_transform(nn.Module):
    '''Implemente color decomposition using 2d-convolutional kernels'''

    def __init__(self, in_channels):
        super().__init__()

        # color transform
        ksize = 1
        stride = 1
        out_channels = in_channels

        f_1 = [0.5, 0.5, 0.5, 0.5];           f_2 = [-0.5, 0.5, 0.5, -0.5]
        f_3 = [0.65, 0.2784, -0.2784, -0.65]; f_4 = [-0.2784, 0.65, -0.65, 0.2784]

        flt = np.stack([f_1,f_2,f_3,f_4], axis=0)
        flt = np.expand_dims(flt, axis=2)
        flt = np.expand_dims(flt,axis=3)
        flt = torch.tensor(flt).float()

        inv_flt = torch.transpose(flt,0,1).clone()

        self.filter_color = nn.Conv2d(in_channels=in_channels, out_channels=out_channels,kernel_size=ksize,stride=stride,bias=False)
        self.inv_filter_color = nn.Conv2d(in_channels=out_channels,out_channels=in_channels,kernel_size=ksize,stride=stride,bias=False)
        # print(self.filter_color)
        
        self.filter_color.weight = nn.Parameter(flt)
        self.inv_filter_color.weight = nn.Parameter(inv_flt)
        # print(self.filter_color.weight)

        for p in self.parameters():
            p.requires_grad=False


        
class EMVDfeat(nn.Module):
    """
    fusionlayers for setting num of layers of Fusion network;
    denoiselayers for setting num of layers of Denoise network;
    refinelayers for setting num of layers of Refinement network;
    """

    def __init__(self, in_channels, out_channels, init_wavelet = 'haar', f_layers=3,f_feats=16,
     d_layers=3, d_feats=16, r_layers=3,r_feats=16,VST=False, ISO=3200):
        super().__init__()

        # Trainanle frequency transform ft. 
        # Penalize the multiplication of ft.dec_filter1d and ft.rec_filter1d
        # ft.dec_filter2d ft.rec_filter2d will be applied in F.conv2d (F.conv_transpose2d) as 'filters'
        self.ft = frequency_transform(in_channels=in_channels,init_wavelet=init_wavelet)

        # Trainable color transform ct.
        # ct.filter_color and ct.inv_filter_color are used for transform and inverse transform.
        self.ct = color_transform(in_channels=in_channels)
        
        self.inchannels = in_channels

        # Three predicting networks: FCNN, DCNN, and RCNN
        self.Fcnn = FCNN(in_channels = 6,  features=f_feats, num_layers=f_layers)
        self.Dcnn = DCNN(in_channels = 21, features=d_feats, num_layers=d_layers)
        self.Rcnn = RCNN(in_channels = 33, features=r_feats, num_layers=r_layers)

        
        self.old_features = None    
        self.fused = None   # Result given by fusion stage, recurrence
        self.denoised = None
        self.NoPF = 1
        self.VST = VST  # apply VST or not
        self.ISO = ISO

        if f_feats == 64:               # for big configuration, avoid skipping denoising network from the start.
            self.epoch_forced_train = 5  # for epochs before self.epoch_forced_train, the refine network doesn't step in, forcing training denoising network
        else:
            self.epoch_forced_train = 0

        # noise map
        self.sigma_noisy = None
        self.sigma_fused = None
        # fusion weights
        self.fusion_weights = None
        self.refine_weights = None


    def noise_map_noisy(self,frame):
        # generate noise map according to sigma = a*y+b
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)

        # self.sigma_noisy = a * frame + b
        if self.VST:
            self.sigma_noisy = a**2/(a+b)*torch.ones(frame.shape, device=frame.device)
        else:
            b = 4*b + 2*a
            a = a/2
            self.sigma_noisy = a*frame+b
        

    def noise_map_fused(self,w1,w2):
        # generate noise map according to sigma = a*y+b
        # modify according to the recursively updated formula
        self.sigma_fused = w1**2 *self.sigma_fused + w2**2 *self.sigma_noisy

    
    def get_rec_init_features(self, noisy):
        # initialize the fake recurrent features by the first noisy frame.
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)
        
        if self.VST:
            noisy = self.variance_stablize(noisy)
        # if torch.any(torch.isnan(noisy)):
        #     print("THERE is NaN!")
        noisy_ct = self.ct.filter_color(noisy)
        noisy_ft = F.conv2d(noisy_ct, self.ft.dec_filter2d, stride=2,groups=self.inchannels)

        if self.VST:
            noise_map_feat = a**2/( a+b)*torch.ones(noisy_ft[:,0:4:4].shape, device=noisy.device )
        else:
            b = 2*a+4*b
            a = a/2
            noise_map_feat = a*noisy_ft[:,0:4:4]+b

        features = []
        features.append( torch.cat( [noisy_ct, F.interpolate(noise_map_feat, scale_factor=2)], dim=1) )
        
        return features

    def set_rec_features(self, features):
        '''
        Set recurrent features, which will be part of input.
        Also set noise map for current recurrent feature.
        '''
        self.sigma_fused = features[0][:,-1:,::2,::2]
        self.old_features = F.conv2d( features[0][:,:-1,:,:],self.ft.dec_filter2d,stride=2,groups=self.inchannels )
        
        
    def get_current_features(self):
        '''
        Update features, which is obtained in current unrolling.
        Noise map of current fusion image is also concatenated, 
        which will also be warped in next unrolling.
        '''
        # Take the fused feature on original domain. (apply inverse haar transform)
        features = []
        # we upscale noise map to fit original scale, also scale of optical flow.
        features.append(  torch.cat([F.conv_transpose2d(self.fused, self.ft.rec_filter2d,stride=2,groups=self.inchannels), \
                                    F.interpolate(self.sigma_fused, scale_factor=2)], dim=1 ) )

        return features

    def set_transforms(self):
        self.ft.dec_filter2d_()
        self.ft.rec_filter2d_()

    def variance_stablize(self, x):
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO) # a, b are for images in [0,1]

        x = (x + 1)/2   # to range [0,1]
        x = torch.sqrt( torch.maximum(a*x+b, torch.zeros_like(x) ) )    # VST, sqrt(ax+b) * 2 / a
        
        a = torch.tensor(a); b = torch.tensor(b)

        return x/torch.sqrt(a+b) *2 -1  # back to [-1,1], x/VST_max * 2 -1

    def inv_variance_stablize(self, x):
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)    # a, b are for images in [0,1]
        a = torch.tensor(a); b = torch.tensor(b)
        x = (x+1)*torch.sqrt(a+b)/a     
        x = a*torch.square(x)/4 - b/a
        '''
        forward VST : sqrt(ax+b)*2/a
        inverse VST : ax^2/4 - b/a
        '''

        return x*2 - 1  # back to [-1,1]


    def forward(self, x, epoch=1, alpha=1,ratioVariance=False):
        # apply VST
        if self.VST:
            x = self.variance_stablize(x)

        # color transform first
        img_ct = self.ct.filter_color(x)
        # frequency transform
        img_ft_sca1 = F.conv2d(img_ct,self.ft.dec_filter2d,stride=2,groups=self.inchannels)

        noisy_LL = img_ft_sca1[:,0:16:4]
        fused_LL = self.old_features[:,0:16:4]
        input_fusion = torch.abs(noisy_LL - fused_LL)
        self.noise_map_noisy(noisy_LL[:,:1])        # noisy subbands, 4 channels

        # fusion
        # input_fusion = torch.cat([input_fusion, self.sigma_noisy],dim=1)
        input_fusion = torch.cat([input_fusion, self.sigma_noisy, self.sigma_fused],dim=1) # add noise map of fusion as input. (together with noise map of noisy)
        fusion_w1, fusion_w2 = self.Fcnn(input_fusion)
        self.fused = fusion_w1*self.old_features + fusion_w2*img_ft_sca1
        self.fusion_weights = fusion_w1
        # denoising
        self.noise_map_fused(fusion_w1,fusion_w2)   # compute noise map recursively
        input_denoise = torch.cat([self.fused, noisy_LL, self.sigma_fused],dim=1)
        self.denoised = self.Dcnn(input_denoise)
        # refinement
        # b,c,h,w = self.sigma_fused.shape
        if epoch>self.epoch_forced_train:
            input_refine = torch.cat([self.fused,self.denoised,self.sigma_fused], dim=1) 
            refine_w1, refine_w2 = self.Rcnn(input_refine, alpha)
        else:
            # print(epoch)
            refine_w1 = 1/2*torch.ones(self.sigma_fused.shape, device=device)
            refine_w2 = refine_w1.clone()
        
        
        refined = refine_w1*self.fused + refine_w2*self.denoised

        self.refine_weights = refine_w2

        # print("weight:")
        # print(self.fusion_weights)
        out = F.conv_transpose2d(refined, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        out = self.ct.inv_filter_color(out)

        # apply inverse VST
        if self.VST:
            out = self.inv_variance_stablize(out)

        return out

    @torch.no_grad()
    def get_fused(self):
        fused_show = F.conv_transpose2d(self.fused, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        fused_show = self.ct.inv_filter_color(fused_show)
        if self.VST:
            return self.inv_variance_stablize(fused_show)
        return fused_show

    def get_denoised(self):
        denoised_show = F.conv_transpose2d(self.denoised, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        denoised_show = self.ct.inv_filter_color(denoised_show)
        if self.VST:    
            return self.inv_variance_stablize(denoised_show)
        return denoised_show

    def get_previous_ta(self):
        temporal_aver = F.conv_transpose2d(self.old_features, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        temporal_aver = self.ct.inv_filter_color(temporal_aver)
        if self.VST:
            return self.inv_variance_stablize(temporal_aver)
        return temporal_aver

    def get_fusion_weights(self):

        return self.fusion_weights.repeat_interleave(4,dim=1)

    def get_refine_weights(self):

        return self.refine_weights.repeat_interleave(4,dim=1)

    def get_fusion_noisemap(self):
        
        return self.sigma_fused.repeat_interleave(4,dim=1)

    def get_noisy_noisemap(self):
        return self.sigma_noisy.repeat_interleave(4,dim=1)




class EMVDfeatbasic(nn.Module):
    """
    fusionlayers for setting num of layers of Fusion network;
    denoiselayers for setting num of layers of Denoise network;
    refinelayers for setting num of layers of Refinement network;
    """

    def __init__(self, in_channels, out_channels, init_wavelet = 'haar', f_layers=3,f_feats=16,
     d_layers=3, d_feats=16, r_layers=3,r_feats=16,VST=False, ISO=3200):
        super().__init__()

        # Trainanle frequency transform ft. 
        # Penalize the multiplication of ft.dec_filter1d and ft.rec_filter1d
        # ft.dec_filter2d ft.rec_filter2d will be applied in F.conv2d (F.conv_transpose2d) as 'filters'
        self.ft = frequency_transform(in_channels=in_channels,init_wavelet=init_wavelet)

        # Trainable color transform ct.
        # ct.filter_color and ct.inv_filter_color are used for transform and inverse transform.
        self.ct = color_transform(in_channels=in_channels)
        
        self.inchannels = in_channels

        # Three predicting networks: FCNN, DCNN, and RCNN
        self.Fcnn = FCNN(in_channels = 5,  features=f_feats, num_layers=f_layers)
        self.Dcnn = DCNN(in_channels = 21, features=d_feats, num_layers=d_layers)
        self.Rcnn = RCNN(in_channels = 33, features=r_feats, num_layers=r_layers)

        
        self.old_features = None    
        self.fused = None   # Result given by fusion stage, recurrence
        self.denoised = None
        self.NoPF = 1
        self.VST = VST  # apply VST or not
        self.ISO = ISO


        # noise map
        self.sigma_noisy = None
        self.sigma_fused = None
        # fusion weights
        self.fusion_weights = None
        self.refine_weights = None


    def noise_map_noisy(self,frame):
        # generate noise map according to sigma = a*y+b
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)

        # self.sigma_noisy = a * frame + b
        if self.VST:
            self.sigma_noisy = a**2/(a+b)*torch.ones(frame.shape, device=frame.device)
        else:
            b = 4*b + 2*a
            a = a/2
            self.sigma_noisy = a*frame+b
        

    def noise_map_fused(self,w1,w2):
        # generate noise map according to sigma = a*y+b
        # modify according to the recursively updated formula
        self.sigma_fused = w1**2 *self.sigma_fused + w2**2 *self.sigma_noisy

    
    def get_rec_init_features(self, noisy):
        # initialize the fake recurrent features by the first noisy frame.
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)
        
        if self.VST:
            noisy = self.variance_stablize(noisy)
        # if torch.any(torch.isnan(noisy)):
        #     print("THERE is NaN!")
        noisy_ct = self.ct.filter_color(noisy)
        noisy_ft = F.conv2d(noisy_ct, self.ft.dec_filter2d, stride=2,groups=self.inchannels)

        if self.VST:
            noise_map_feat = a**2/( a+b)*torch.ones(noisy_ft[:,0:4:4].shape, device=noisy_ct.device )
        else:
            b = 2*a+4*b
            a = a/2
            noise_map_feat = a*noisy_ft[:,0:4:4]+b

        features = []
        features.append( torch.cat( [noisy_ct, F.interpolate(noise_map_feat, scale_factor=2)], dim=1) )
        
        return features

    def set_rec_features(self, features):
        '''
        Set recurrent features, which will be part of input.
        Also set noise map for current recurrent feature.
        '''
        self.sigma_fused = features[0][:,-1:,::2,::2]
        self.old_features = F.conv2d( features[0][:,:-1,:,:],self.ft.dec_filter2d,stride=2,groups=self.inchannels )
        
        
    def get_current_features(self):
        '''
        Update features, which is obtained in current unrolling.
        Noise map of current fusion image is also concatenated, 
        which will also be warped in next unrolling.
        '''
        # Take the fused feature on original domain. (apply inverse haar transform)
        features = []
        # we upscale noise map to fit original scale, also scale of optical flow.
        features.append(  torch.cat([F.conv_transpose2d(self.fused, self.ft.rec_filter2d,stride=2,groups=self.inchannels), \
                                    F.interpolate(self.sigma_fused, scale_factor=2)], dim=1 ) )

        return features

    def set_transforms(self):
        self.ft.dec_filter2d_()
        self.ft.rec_filter2d_()

    def variance_stablize(self, x):
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO) # a, b are for images in [0,1]

        x = (x + 1)/2   # to range [0,1]
        x = torch.sqrt( torch.maximum(a*x+b, torch.zeros_like(x) ) )    # VST, sqrt(ax+b) * 2 / a
        
        a = torch.tensor(a); b = torch.tensor(b)

        return x/torch.sqrt(a+b) *2 -1  # back to [-1,1], x/VST_max * 2 -1

    def inv_variance_stablize(self, x):
        from util.addNoise import noiseModel
        a, b = noiseModel(self.ISO)    # a, b are for images in [0,1]
        a = torch.tensor(a); b = torch.tensor(b)
        x = (x+1)*torch.sqrt(a+b)/a     
        x = a*torch.square(x)/4 - b/a
        '''
        forward VST : sqrt(ax+b)*2/a
        inverse VST : ax^2/4 - b/a
        '''

        return x*2 - 1  # back to [-1,1]

    def fusion_weight(self, w):
        ratio = self.sigma_noisy / (self.sigma_noisy+self.sigma_fused)
        w = w * ratio   # weight assigned for temporal fusion
        return w, 1-w


    def forward(self, x, ratioVariance=False):
        # apply VST
        if self.VST:
            x = self.variance_stablize(x)

        # color transform first
        img_ct = self.ct.filter_color(x)
        # frequency transform
        img_ft_sca1 = F.conv2d(img_ct,self.ft.dec_filter2d,stride=2,groups=self.inchannels)

        noisy_LL = img_ft_sca1[:,0:16:4]
        fused_LL = self.old_features[:,0:16:4]
        input_fusion = torch.abs(noisy_LL - fused_LL)
        self.noise_map_noisy(noisy_LL[:,:1])        # noisy subbands, 4 channels

        # fusion
        input_fusion = torch.cat([input_fusion, self.sigma_noisy],dim=1)
        if ratioVariance:
            # print('Decoupling occlusion detection and weight estimation')
            fusion_w, _ = self.Fcnn(input_fusion)
            fusion_w1,fusion_w2 = self.fusion_weight(fusion_w)
        else:
            fusion_w1, fusion_w2 = self.Fcnn(input_fusion)
        self.fused = fusion_w1*self.old_features + fusion_w2*img_ft_sca1
        self.fusion_weights = fusion_w1
        # denoising
        self.noise_map_fused(fusion_w1,fusion_w2)   # compute noise map recursively
        input_denoise = torch.cat([self.fused, noisy_LL, self.sigma_fused],dim=1)
        self.denoised = self.Dcnn(input_denoise)
        # refinement     
        input_refine = torch.cat([self.fused,self.denoised,self.sigma_fused], dim=1) 
        refine_w1, refine_w2 = self.Rcnn(input_refine)        
        refined = refine_w1*self.fused + refine_w2*self.denoised

        self.refine_weights = refine_w2

        out = F.conv_transpose2d(refined, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        out = self.ct.inv_filter_color(out)

        # apply inverse VST
        if self.VST:
            out = self.inv_variance_stablize(out)

        return out

    @torch.no_grad()
    def get_fused(self):
        fused_show = F.conv_transpose2d(self.fused, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        fused_show = self.ct.inv_filter_color(fused_show)
        if self.VST:
            return self.inv_variance_stablize(fused_show)
        return fused_show

    def get_denoised(self):
        denoised_show = F.conv_transpose2d(self.denoised, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        denoised_show = self.ct.inv_filter_color(denoised_show)
        if self.VST:    
            return self.inv_variance_stablize(denoised_show)
        return denoised_show

    def get_previous_ta(self):
        temporal_aver = F.conv_transpose2d(self.old_features, self.ft.rec_filter2d,stride=2,groups=self.inchannels)
        temporal_aver = self.ct.inv_filter_color(temporal_aver)
        if self.VST:
            return self.inv_variance_stablize(temporal_aver)
        return temporal_aver

    def get_fusion_weights(self):

        return self.fusion_weights.repeat_interleave(4,dim=1)

    def get_refine_weights(self):

        return self.refine_weights.repeat_interleave(4,dim=1)

    def get_fusion_noisemap(self):
        
        return self.sigma_fused.repeat_interleave(4,dim=1)

    def get_noisy_noisemap(self):
        return self.sigma_noisy.repeat_interleave(4,dim=1)



