import argparse
import os
from util import util
import torch
import models
import data


class BaseOptions():
    """This class defines options used during both training and test time.

    It also implements several helper functions such as parsing, printing, and saving the options.
    It also gathers additional options defined in <modify_commandline_options> functions in both dataset class and model class.
    """

    def __init__(self):
        """Reset the class; indicates the class hasn't been initailized"""
        self.initialized = False

    def initialize(self, parser):
        """Define the common options that are used in both training and test."""

        # basic parameters
        parser.add_argument('--gpu_ids', type=str, default='0', help='gpu ids: e.g. 0  0,1,2, 0,2. use -1 for CPU')
        parser.add_argument('--checkpoints_dir', type=str, default='./checkpoints', help='models are saved here')
        # model parameters
        parser.add_argument('--model', type=str, default='recurrent', help='chooses which model to use. [ basic | recurrent | dummy ]')
        parser.add_argument('--input_nc', type=int, default=4, help='# of input image channels: 3 for RGB and 1 for grayscale')
        parser.add_argument('--output_nc', type=int, default=4, help='# of output image channels: 3 for RGB and 1 for grayscale')
        parser.add_argument('--ngf', type=int, default=64, help='# of gen filters in the last conv layer')
        parser.add_argument('--ndf', type=int, default=64, help='# of discrim filters in the first conv layer')
        parser.add_argument('--netDenoiser', type=str, default='resnet_9', help='specify generator architecture [resnet_N | resnet_9blocks | resnet_6blocks | unet_256 | unet_128]')
        parser.add_argument('--norm', type=str, default='instance', help='instance normalization or batch normalization [instance | batch | none]')
        parser.add_argument('--init_type', type=str, default='kaiming', help='network initialization [normal | xavier | kaiming | orthogonal]')
        parser.add_argument('--init_gain', type=float, default=0.02, help='scaling factor for normal, xavier and orthogonal.')
        parser.add_argument('--no_dropout', action='store_true', help='no dropout for the generator')
        
        # dataset parameters
        parser.add_argument('--dataroot', type=str, default='./datasets/train_dataset', required=False, help='Root folder in which to look for noisy and groundtruth.')
        parser.add_argument('--nFolder', type=str, default='noisy', required=False, help='Subfolder of "dataroot" in which to find the noisy data.')
        parser.add_argument('--gtFolder', type=str, default='gt', required=False, help='Subfolder of "dataroot" in which to find the groundtruth data.')
        parser.add_argument('--wFolder', type=str, default='warped', required=False, help='Subfolder of "dataroot" in which to find the warped data.')
        parser.add_argument('--mFolder', type=str, default='masked', required=False, help='Subfolder of "dataroot" in which to find the masked data.')
        parser.add_argument('--flowFolder', type=str, default='flow', required=False, help='Subfolder of "dataroot" in which to find the flow data.')
        parser.add_argument('--rawFolder', type=str, default=None, required=False, help='Subfolder of "dataroot" in which to find the raw data. Can be None.')
        parser.add_argument('--data_nocheck', default=False, action='store_true', help='it bypasses checking if all data has been properly generated.')
        parser.add_argument('--warp', action='store_true', default=False, help='Specifies if all 2d patches should be warped to fit the last one.')
        parser.add_argument('--warp_method', type=str, default='tvl1', required=False, help='Warping method [tvl1, ica]')
        parser.add_argument('--warp_in_torch',  action='store_true', default=True, help='Tells if all warps should be done with the same Torch function as in training.')
        parser.add_argument('--model_input_gen', type=str, default='noisyinputs', help='chooses which model to use. [ lastnoisy | noisyinputs ]')
        parser.add_argument('--mask', action='store_true', default=False, help='Uses a precomputed mask of non corresponding pixels between patches.')
        parser.add_argument('--mask_value', type=float, default=-1024, help='Mask value for zero. Ones are left unchanged.')
        parser.add_argument('--mask_on_chan', action='store_true', default=False, help='Specifies if the mask is passed as a channel. If false then it will set the values of masked pixels to mask_value.')
        parser.add_argument('--videos',  type=str, default=None, help='which videos to use.') # works for inferaxel and axel4rec

        parser.add_argument('--dataset_mode', type=str, default='axel4rec', help='chooses how datasets are loaded. [ axel | inferaxel | axel4rec | dummy ]')
        parser.add_argument('--serial_batches', default=False, action='store_true', help='if true, takes images in order to make batches, otherwise takes them randomly')
        parser.add_argument('--pin_memory', default=True, action='store_true', help='if true, it pins memory whenever it could be pinned.')
        parser.add_argument('--non_blocking', default=True, action='store_true', help='if true, it non-blocks memory copying whenever it could be.')
        parser.add_argument('--num_threads', default=4, type=int, help='# threads for loading data')
        parser.add_argument('--batch_size', type=int, default=26, help='input batch size')
        parser.add_argument('--Nmomentum', type=float, default=0.1, help='momentum of BatchNormalisation or InstanceNormalisation')
        parser.add_argument('--max_dataset_size', type=int, default=float("inf"), help='Maximum number of samples allowed per dataset. If the dataset directory contains more than max_dataset_size, only a subset is loaded.')
        parser.add_argument('--preprocess', type=str, default='resize_and_crop', help='scaling and cropping of images at load time [resize_and_crop | crop | scale_width | scale_width_and_crop | none]')
        parser.add_argument('--no_flip', action='store_true', help='if specified, do not flip the images for data augmentation')
        parser.add_argument('--display_winsize', type=int, default=256, help='display window size for both visdom and HTML')
        parser.add_argument('--load_size', type=int, default=286, help='scale images to this size (might not be used by the dataset loader).')
        parser.add_argument('--patch_width',  type=int, default=96, help='patches of this size WxW')
        parser.add_argument('--patch_depth',  type=int, default=2, help='a sequence of consecutive #depth patches')
        parser.add_argument('--patch_stride',  type=int, default=3, help='temporal and spatial stride (or step) to form the 3D patch')
        # additional parameters
        parser.add_argument('--epoch', type=str, default='latest_val', help='which epoch to load? set to latest to use latest cached model')
        parser.add_argument('--load_iter', type=int, default='0', help='which iteration to load? if load_iter > 0, the code will load models by iter_[load_iter]; otherwise, the code will load models by [epoch]')
        parser.add_argument('--verbose', action='store_true', help='if specified, print more debugging information')
        parser.add_argument('--suffix', default='', type=str, help='customized suffix: opt.name = opt.name + suffix')
        parser.add_argument('--no_eval', action='store_true', default=False, help='Dont let model.eval() set layers like batchnorm and dropout.')
        parser.add_argument('--visuals', type=str, default='debug', help='[debug,denoiser,none]')

        parser.add_argument('--ratioVariance', action='store_true', help='if true, adjust the fusion weight by Variacne of fusion and noisy.')

        self.initialized = True
        return parser

    def gather_options(self):
        """Initialize our parser with basic options(only once).
        Add additional model-specific and dataset-specific options.
        These options are defined in the <modify_commandline_options> function
        in model and dataset classes.
        """
        if not self.initialized:  # check if it has been initialized
            parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter)
            parser = self.initialize(parser)

        # get the basic options
        opt, _ = parser.parse_known_args()

        # modify model-related parser options
        model_name = opt.model
        model_option_setter = models.get_option_setter(model_name)
        parser = model_option_setter(parser, self.isTrain)
        opt, _ = parser.parse_known_args()  # parse again with new defaults

        # modify dataset-related parser options
        dataset_name = opt.dataset_mode
        dataset_option_setter = data.get_option_setter(dataset_name)
        parser = dataset_option_setter(parser, self.isTrain)

        # save and return the parser
        self.parser = parser
        return parser.parse_args()

    def print_options(self, opt):
        """Print and save options

        It will print both current options and default values(if different).
        It will save options into a text file / [checkpoints_dir] / opt.txt
        """
        message = ''
        message += '----------------- Options ---------------\n'
        for k, v in sorted(vars(opt).items()):
            comment = ''
            default = self.parser.get_default(k)
            if v != default:
                comment = '\t[default: %s]' % str(default)
            message += '{:>25}: {:<30}{}\n'.format(str(k), str(v), comment)
        message += '----------------- End -------------------'
        print(message)

        # save to the disk
        expr_dir = os.path.join(opt.checkpoints_dir, opt.name)
        util.mkdirs(expr_dir)
        if self.isTrain:
            file_name = os.path.join(expr_dir, 'opt_train.txt')
        else:
            file_name = os.path.join(expr_dir, 'opt_test.txt')
        with open(file_name, 'wt') as opt_file:
            opt_file.write(message)
            opt_file.write('\n')
    
    @staticmethod
    def update_name(opt):
        warpstr = '-warp' if opt.warp else ''
        maskstr = ''
        if opt.mask and not opt.mask_on_chan:
            maskstr = '-ipmask' 
        elif opt.mask and opt.mask_on_chan:
            maskstr = '-acmask'
        rawstr = '' if opt.rawFolder in ["\noraw", " ", None] else '-raw'
        suffixstr = "-" + opt.suffix if not opt.suffix == '' else ""
        opt.name = "%s-%s-%s%s%s%s-i%do%d%s" % (opt.model, opt.netDenoiser,
                                              opt.model_input_gen, warpstr, maskstr, rawstr, opt.input_nc, opt.output_nc, suffixstr)
        return opt

    def parse(self):
        """Parse our options, create checkpoints directory suffix, and set up gpu device."""
        opt = self.gather_options()
        opt.isTrain = self.isTrain   # train or test

        opt = self.update_name(opt)

        self.print_options(opt)

        # set gpu ids
        str_ids = opt.gpu_ids.split(',')
        opt.gpu_ids = []
        for str_id in str_ids:
            id = int(str_id)
            if id >= 0:
                opt.gpu_ids.append(id)
        if len(opt.gpu_ids) > 0:
            torch.cuda.set_device(opt.gpu_ids[0])

        self.opt = opt
        return self.opt
